@@ -5,10 +5,14 @@
  * for all jukebox functionality.
  */
 
+/* MSVC does not have these */
+#ifndef _MSC_VER
 #include "config.h"
+#include <unistd.h>
+#endif
+
 #include <stdio.h>
 #include <stdlib.h>
-#include <unistd.h>
 #include <fcntl.h>
 #include <string.h>
 #include <errno.h>
